"""Useful defualt settings that you might want to use in all your projects."""
import re


IGNORABLE_404_URLS = [
    re.compile(r'\..*$'),
    re.compile(r'^/media/'),
    re.compile(r'^/static/'),
]



"""URLs for the test app."""
from django.conf.urls import patterns, url
from django.http import HttpResponse
from django.views.generic import View

from django_libs import views

View.get = lambda req, *args, **kwargs: HttpResponse('SUCCESS!')
authed_view = View.as_view()
authed_view_kwargs = {'authed': True}
anonymous_view = View.as_view()
anonymous_view_kwargs = {'anonymous': True}


urlpatterns = patterns(
    '',
    url(r'^$', views.HybridView.as_view(
        authed_view=authed_view,
        authed_view_kwargs=authed_view_kwargs,
        anonymous_view=anonymous_view,
        anonymous_view_kwargs=anonymous_view_kwargs,
        ), name='dummy_hybrid'),
    url(r'^update-session/$', views.UpdateSessionAJAXView,
        name='update_session'),
    url(r'^update-cookie/$', views.UpdateCookieAJAXView,
        name='update_cookie'),
    url(r'^prototype/(?P<template_path>.*)$',
        views.RapidPrototypingView.as_view(),
        name='prototype')
)



"""Tests for the view classes of ``django-libs``."""
from django.contrib.sites.models import Site
from django.contrib.comments.models import Comment
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase
from django.views.generic import TemplateView, View

from .. import views
from .factories import UserFactory
from .mixins import ViewTestMixin, ViewRequestFactoryTestMixin


class HybridViewTestCase(ViewTestMixin, TestCase):
    """Tests for the ``HybridView`` view class."""
    longMessage = True

    def setUp(self):
        self.user = UserFactory()

        self.authed_view = TemplateView.as_view(template_name='base.html')
        self.authed_view_kwargs = {'authed': True}
        self.anonymous_view = TemplateView.as_view(template_name='base.html')
        self.anonymous_view_kwargs = {'anonymous': True}
        self.other_anonymous_view = View.as_view()
        self.view_kwargs = {
            'authed_view': self.authed_view,
            'authed_view_kwargs': self.authed_view_kwargs,
            'anonymous_view': self.anonymous_view,
            'anonymous_view_kwargs': self.anonymous_view_kwargs}

    def get_view_name(self):
        return self.view_name

    def test_view(self):
        self.view_name = 'dummy_hybrid'
        self.should_be_callable_when_anonymous()

        bad_kwargs = self.view_kwargs.copy()
        bad_kwargs.update({'post': 'this should not be defined here'})
        self.assertRaises(TypeError, views.HybridView.as_view, **bad_kwargs)

        bad_kwargs = self.view_kwargs.copy()
        bad_kwargs.update({'wrongattr': 'this is not defined on the view'})
        self.assertRaises(TypeError, views.HybridView.as_view, **bad_kwargs)

        self.should_be_callable_when_authenticated(self.user)


class PaginatedCommentAJAXViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``PaginatedCommentAJAXView`` view class."""
    view_class = views.PaginatedCommentAJAXView

    def setUp(self):
        self.user = UserFactory()
        self.comment = Comment.objects.create(
            content_type=ContentType.objects.get(name='user'),
            object_pk=self.user.pk,
            comment='foobar',
            site=Site.objects.create(),
        )

    def test_view(self):
        self.is_not_callable()
        self.is_callable(ajax=True)


class RapidPrototypingViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``RapidPrototypingView`` view class."""
    longMessage = True

    def setUp(self):
        self.view_class = views.RapidPrototypingView

    def get_view_name(self):
        return 'prototype'

    def get_view_kwargs(self):
        return {'template_path': 'django_libs/analytics.html'}

    def test_view(self):
        self.is_callable()


class UpdateSessionAJAXViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``UpdateSessionAJAXView`` view class."""
    longMessage = True

    def setUp(self):
        self.view_class = views.UpdateSessionAJAXView

    def get_view_name(self):
        return 'update_session'

    def test_view(self):
        self.is_forbidden()
        data = {'session_name': 'foo', 'session_value': 'bar'}
        resp = self.is_postable(ajax=True, data=data)
        self.assertEqual(resp.content, 'done')


class UpdateCookieAJAXViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``UpdateCookieAJAXView`` view class."""
    longMessage = True

    def setUp(self):
        self.view_class = views.UpdateCookieAJAXView

    def get_view_name(self):
        return 'update_cookie'

    def test_view(self):
        self.is_forbidden()
        data = {'cookie_key': 'foo', 'cookie_value': 'bar'}
        resp = self.is_postable(ajax=True, data=data)
        self.assertEqual(resp.content, 'done')



"""Custom S3 storage backends to store files in subfolders."""
from django.core.files.storage import get_storage_class

from storages.backends.s3boto import S3BotoStorage


class CachedS3BotoStorage(S3BotoStorage):
    def __init__(self, *args, **kwargs):
        super(CachedS3BotoStorage, self).__init__(*args, **kwargs)
        self.local_storage = get_storage_class(
            'compressor.storage.CompressorFileStorage')()

    def save(self, name, content):
        name = super(CachedS3BotoStorage, self).save(name, content)
        self.local_storage._save(name, content)
        return name


CompressorS3BotoStorage = lambda: CachedS3BotoStorage(location='compressor')
MediaRootS3BotoStorage = lambda: S3BotoStorage(location='media')



"""Tests for the model mixins of ``django_libs``."""
from mock import Mock

from django.test import TestCase

from .test_app.factories import DummyProfileTranslationFactory
from .test_app.models import DummyProfile


class SimpleTranslationMixinTestCase(TestCase):
    """Tests for the ``SimpleTranslationMixin`` mixin."""
    longMessage = True

    def setUp(self):
        self.dummyprofile_trans = DummyProfileTranslationFactory()
        self.dummyprofile = self.dummyprofile_trans.dummyprofile

    def test_mixin(self):
        self.assertEqual(self.dummyprofile.get_translation(),
                         self.dummyprofile_trans)


class SimpleTranslationPublishedManagerTestCase(TestCase):
    """Tests for the ``SimpleTranslationPublishedManager`` manager."""
    longMessage = True

    def setUp(self):
        DummyProfileTranslationFactory()
        DummyProfileTranslationFactory(is_published=False)
        DummyProfileTranslationFactory(language='de')
        DummyProfileTranslationFactory(language='de', is_published=False)

    def test_manager(self):
        request = Mock(LANGUAGE_CODE='en')
        self.assertEqual(DummyProfile.objects.published(request).count(), 1,
                         msg='There should be one published english dummy.')

        request = Mock(LANGUAGE_CODE='de')
        self.assertEqual(DummyProfile.objects.published(request).count(), 1,
                         msg='There should be one published german dummy.')

        request = Mock(LANGUAGE_CODE=None)
        self.assertEqual(DummyProfile.objects.published(request).count(), 0,
                         msg=(
                             'There should be no published dummy, if no'
                             ' language is set.'))



"""Tests for the view mixins of ``django-libs``."""
from django.test import TestCase
from django.test.client import RequestFactory
from django.views.generic import TemplateView

from django_libs.views_mixins import AjaxResponseMixin


class DummyView(AjaxResponseMixin, TemplateView):
    """Just a test view."""
    template_name = "test_template.html"


class AjaxResponseMixinTestCase(TestCase):
    longMessage = True

    def setUp(self):
        self.view = DummyView()

    def test_mixin(self):
        """Test for the ``AjaxResponseMixin`` class."""
        extra = {'HTTP_X_REQUESTED_WITH': 'XMLHttpRequest'}
        req = RequestFactory().get('/', **extra)
        self.view.request = req
        self.assertEqual(self.view.get_template_names(),
                         ['ajax_test_template.html'],
                         msg='Got the wrong template name.')



"""Useful decorators for Django projects."""
from functools import wraps
import re

from django.http import Http404
from django.contrib.auth import authenticate, login
from django.contrib.auth.models import User
from lockfile import FileLock, AlreadyLocked, LockTimeout


def lockfile(lockfile_name, lock_wait_timeout=-1):
    """
    Only runs the method if the lockfile is not acquired.

    You should create a setting ``LOCKFILE_PATH`` which points to
    ``/home/username/tmp/``.

    In your management command, use it like so::

        LOCKFILE = os.path.join(
            settings.LOCKFILE_FOLDER, 'command_name')

        class Command(NoArgsCommand):
            @lockfile(LOCKFILE)
            def handle_noargs(self, **options):
                # your command here

    :lockfile_name: A unique name for a lockfile that belongs to the wrapped
      method.
    :lock_wait_timeout: Seconds to wait if lockfile is acquired. If ``-1`` we
      will not wait and just quit.

    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            lock = FileLock(lockfile_name)
            try:
                lock.acquire(lock_wait_timeout)
            except AlreadyLocked:
                return
            except LockTimeout:
                return
            try:
                result = func(*args, **kwargs)
            finally:
                lock.release()
            return result

        return wrapper
    return decorator


def get_username(identifier):
    """Checks if a string is a email adress or not."""
    pattern = re.compile('.+@\w+\..+')
    if pattern.match(identifier):
        try:
            user = User.objects.get(email=identifier)
        except:
            raise Http404
        else:
            return user.username
    else:
        return identifier


def http_auth(func):
    @wraps(func)
    def _decorator(request, *args, **kwargs):
        """Decorator to handle http authorizations."""
        if request.user.is_authenticated():
            return func(request, *args, **kwargs)
        if 'HTTP_AUTHORIZATION' in request.META.keys():
            authmeth, auth = request.META['HTTP_AUTHORIZATION'].split(' ', 1)
            if authmeth.lower() == 'basic':
                auth = auth.strip().decode('base64')
                identifier, password = auth.split(':', 1)
                username = get_username(identifier)
                user = authenticate(username=username, password=password)
                if user:
                    login(request, user)
                    return func(request, *args, **kwargs)
        raise Http404
    return _decorator



"""URLs to run the tests."""
from django.conf.urls import patterns, include, url
from django.contrib import admin
from django.http import HttpResponse

admin.autodiscover()


urlpatterns = patterns(
    '',
    url(r'^index/test/$', lambda x: HttpResponse('Success'), name='index'),
    url(r'^admin-.+/', include(admin.site.urls)),
    url(r'^', include('test_app.urls')),
)



"""Just some factories for the test app."""
import factory

from ..factories import UserFactory
from models import DummyProfile, DummyProfileTranslation


class DummyProfileFactory(factory.DjangoModelFactory):
    """Factory for the ``DummyProfile`` model."""
    FACTORY_FOR = DummyProfile

    user = factory.SubFactory(UserFactory)
    dummy_field = factory.Sequence(lambda n: 'dummyfield{}'.format(n))


class DummyProfileTranslationFactory(factory.DjangoModelFactory):
    """Factory for the ``DummyProfileTranslation`` model."""
    FACTORY_FOR = DummyProfileTranslation

    dummy_translation = factory.Sequence(lambda n: 'trans {}'.format(n))
    dummyprofile = factory.SubFactory(DummyProfileFactory)



"""Custom middleware for Django 1.5 projects."""
import hashlib

from django import http
from django.conf import settings
from django.core.mail import mail_managers
from django.middleware.common import (
    CommonMiddleware,
    _is_ignorable_404,
    _is_internal_request,
)


class CustomCommonMiddleware(CommonMiddleware):
    """Adds the current user to the 405 email."""
    def process_response(self, request, response):
        "Send broken link emails and calculate the Etag, if needed."
        if response.status_code == 404:
            if settings.SEND_BROKEN_LINK_EMAILS and not settings.DEBUG:
                # If the referrer was from an internal link or a
                # non-search-engine site, send a note to the managers.
                domain = request.get_host()
                referer = request.META.get('HTTP_REFERER')
                is_internal = _is_internal_request(domain, referer)
                path = request.get_full_path()
                if (referer
                        and not _is_ignorable_404(path)
                        and (is_internal or '?' not in referer)):
                    ua = request.META.get('HTTP_USER_AGENT', '<none>')
                    ip = request.META.get('REMOTE_ADDR', '<none>')

                    user = None
                    if request.user and hasattr(request.user, 'email'):
                        user = request.user.email

                    content = (
                        "Referrer: %s\n"
                        "Requested URL: %s\n"
                        "User agent: %s\n"
                        "IP address: %s\n"
                        "User: %s\n"
                    ) % (referer, request.get_full_path(), ua, ip, user)
                    internal = is_internal and 'INTERNAL ' or ''
                    mail_managers(
                        "Broken %slink on %s" % (internal, domain),
                        content,
                        fail_silently=True,
                    )
                return response

        # Use ETags, if requested.
        if settings.USE_ETAGS:
            if response.has_header('ETag'):
                etag = response['ETag']
            elif response.streaming:
                etag = None
            else:
                etag = '"%s"' % hashlib.md5(response.content).hexdigest()
            if etag is not None:
                if (200 <= response.status_code < 300
                        and request.META.get('HTTP_IF_NONE_MATCH') == etag):
                    cookies = response.cookies
                    response = http.HttpResponseNotModified()
                    response.cookies = cookies
                else:
                    response['ETag'] = etag

        return response



"""Additions for the django admin."""
from django.utils.translation import ugettext_lazy as _


class MultilingualPublishMixin(object):
    """Mixin to provide common methods for multilingual object admins."""

    def is_published(self, obj):
        languages = ''
        for trans in obj.translations:
            if trans.is_published:
                if languages == '':
                    languages = trans.language
                else:
                    languages += ', {0}'.format(trans.language)
        if languages == '':
            return _('Not published')
        return languages
    is_published.short_description = _('Is published')



"""Useful mixins for class based views."""
import json
import os

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseRedirect
from django.views.generic import DetailView


class AccessMixin(object):
    """Mixin to controls access to the view based on a setting."""
    access_mixin_setting_name = None

    def dispatch(self, request, *args, **kwargs):
        # Check, if user needs to be logged in
        if self.access_mixin_setting_name is None:
            raise Exception(
                'Please set `access_mixin_setting_name` on the view that'
                ' inherits the AccessMixin')
        if getattr(settings, self.access_mixin_setting_name, False):
            return super(AccessMixin, self).dispatch(
                request, *args, **kwargs)
        return login_required(super(AccessMixin, self).dispatch)(
            request, *args, **kwargs)


class AjaxResponseMixin(object):
    """
    A mixin that prepends `ajax_` to the template name when it is an ajax call.

    This gives you the chance to return partial templates when it is an ajax
    call, so you can render the output inside of a modal, for example.

    """
    ajax_template_prefix = 'ajax_'

    def get_template_names(self):
        names = super(AjaxResponseMixin, self).get_template_names()
        if self.request.is_ajax():
            count = 0
            for name in names:
                filename_split = list(os.path.split(name))
                old_filename = filename_split[-1]
                new_filename = '{0}{1}'.format(
                    self.ajax_template_prefix, old_filename)
                filename_split[-1] = new_filename
                names[count] = os.path.join(*filename_split)
                count += 1
        return names


class DetailViewWithPostAction(DetailView):
    """
    Generic class based view to handle custom post actions in a DetailView.

    When you derive from this class, your buttons need to be called
    `post_actionname` and you have to implement action handlers with the
    name `post_actionname` and url retrievers with the name
    `get_success_url_post_actionname`.

    If all actions should have the same success url, you can also implement
    `get_success_url`, which will be used as a fallback in case that no
    specific url retrieve has been implemented.

    """
    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        for key in self.request.POST.keys():
            if key.startswith('post_'):
                getattr(self, key)()
                break
        success_url_handler = getattr(self, 'get_success_url_%s' % key, False)
        if not success_url_handler:
            success_url_handler = getattr(self, 'get_success_url')
        success_url = success_url_handler()
        return HttpResponseRedirect(success_url)


class JSONResponseMixin(object):
    """
    A mixin that can be used to render a JSON response.

    Taken from here: https://docs.djangoproject.com/en/dev/topics/
    class-based-views/#more-than-just-html

    """
    response_class = HttpResponse

    def render_to_response(self, context, **response_kwargs):
        """
        Returns a JSON response, transforming 'context' to make the payload.

        """
        response_kwargs['content_type'] = 'application/json'
        return self.response_class(
            self.convert_context_to_json(context),
            **response_kwargs
        )

    def convert_context_to_json(self, context):
        """
        Convert the context dictionary into a JSON object.

        If your context has complex Django objects, you need to override this
        method and make sure that the context gets transformed into something
        that ``json.dumps`` can handle.

        """
        return json.dumps(context)



"""Useful context  processors for your projects."""
from django.conf import settings


def analytics(request):
    """Adds the setting ANALYTICS_TRACKING_ID to the template context."""
    return {
        'ANALYTICS_TRACKING_ID': getattr(
            settings, 'ANALYTICS_TRACKING_ID', 'UA-XXXXXXX-XX'),
    }



"""
Factories that are common to most Django apps.

The factories in this module shall help to create test fixtures for models that
are global to all Django projects and could be shared by tests of specialized
apps.

For example each app will need to create a user, therefore this module shall
provide facilities for user generation.

"""
from io import BytesIO

from django.contrib.auth.models import User
from django.core.files.uploadedfile import SimpleUploadedFile

from PIL import Image
from hashlib import md5
import factory


class HvadFactoryMixin(object):
    """
    Overrides ``_create`` and takes care of creating a translation.

    """
    @classmethod
    def _create(cls, target_class, *args, **kwargs):
        obj = target_class(*args, **kwargs)

        # Factory boy and hvad behave a bit weird. When getting the object,
        # obj.some_translatable_field is actually set although no translation
        # object exists, yet. We have to cache the translatable values ...
        cached_values = {}
        for field in obj._translated_field_names:
            if field in ['id', 'master', 'language_code']:
                continue
            cached_values[field] = getattr(obj, field)

        # ... because when calling translate, the translatable values will be
        # lost on the obj ...
        obj.translate(obj.language_code)
        for field in obj._translated_field_names:
            if field in ['id', 'master', 'language_code']:
                continue
            # ... so here we will put them back on the object, this time they
            # will be saved on the translatable object.
            setattr(obj, field, cached_values[field])

        obj.save()
        return obj


class UploadedImageFactory(object):
    """Creates an uploaded image for testing."""
    def __new__(cls, **kwargs):
        return cls._create_image(**kwargs)

    @classmethod
    def _create_image(self, image_format='JPEG', filename='img.jpg'):
        thumb = Image.new('RGB', (100, 100), 'blue')
        thumb_io = BytesIO()
        thumb.save(thumb_io, format=image_format)
        self._image = SimpleUploadedFile(filename, thumb_io.getvalue())
        return self._image


class SimpleTranslationMixin(object):
    """
    Adds a ``_prepare`` method that takes care of creating a translation.

    """

    @staticmethod
    def _get_translation_factory_and_field(self):
        """
        Should return a tuple of (TranslationFactory, 'fieldname').

        ``TranslationFactory`` is the factory class that can create translation
        objects for this objects.

        ``fieldname`` is the name of the FK on the translation class that
        points back to this object.

        """
        raise NotImplementedError()

    @classmethod
    def _prepare(cls, create, **kwargs):
        """
        Creates a ``PersonTranslation`` for this ``Person``.

        Note that we will only create a translation if you create a new object
        instead of just building it, because otherwise this object has no PK
        and cannot be used to instantiate the translation.

        """
        language = kwargs.pop('language', 'en')
        obj = super(SimpleTranslationMixin, cls)._prepare(create, **kwargs)
        if create:
            if language:
                translation_factory, fk_field = \
                    cls._get_translation_factory_and_field()
                kwargs_ = {
                    fk_field: obj,
                    'language': language,
                }
                translation_factory(**kwargs_)
        return obj


class UserFactory(factory.DjangoModelFactory):
    """
    Creates a new ``User`` object.

    We use ``django-registration-email`` which allows users to sign in with
    their email instead of a username. Since the username field is too short
    for most emails, we don't really use the username field at all and just
    store a md5 hash in there.

    Username will be a random 30 character md5 value.
    Email will be ``userN@example.com`` with ``N`` being a counter.
    Password will be ``test123`` by default.

    """
    FACTORY_FOR = User

    username = factory.Sequence(lambda n: md5(str(n)).hexdigest()[0:30])
    email = factory.Sequence(lambda n: 'user{0}@example.com'.format(n))

    @classmethod
    def _prepare(cls, create, **kwargs):
        password = 'test123'
        if 'password' in kwargs:
            password = kwargs.pop('password')
        user = super(UserFactory, cls)._prepare(create, **kwargs)
        user.set_password(password)
        if create:
            user.save()
        return user



"""Default settings for the CustomRaven404CatchMiddleware"""
RAVEN_IGNORABLE_USER_AGENTS = [
    r'AhrefsBot',
    r'EasouSpider',
    r'FacebookBot',
    r'Feedfetcher-Google',
    r'Googlebot',
    r'Mail.RU_Bot',
    r'Test Certificate Info',
    r'Twitterbot',
    r'Yahoo! Slurp',
    r'YandexBot',
    r'bingbot',
]



"""Custom email backend for testing the project."""
import re

from django.core.mail.backends.smtp import EmailBackend as SmtpEmailBackend
from django.core.mail.message import sanitize_address

from . import default_settings as settings


class EmailBackend(SmtpEmailBackend):
    """
    Email backend that sends all emails to a defined address, no matter what
    the recipient really is.

    In order to use it, set this in your local_settings.py::

        EMAIL_BACKEND = 'django_libs.test_email_backend.EmailBackend'
        TEST_EMAIL_BACKEND_RECIPIENTS = (
            ('Name', 'email@gmail.com'),
        )

    """
    def _send(self, email_message):
        """A helper method that does the actual sending."""
        if not email_message.recipients() or \
                not settings.TEST_EMAIL_BACKEND_RECIPIENTS:
            return False
        from_email = sanitize_address(
            email_message.from_email, email_message.encoding)
        recipients = [sanitize_address(addr, email_message.encoding)
                      for name, addr in settings.TEST_EMAIL_BACKEND_RECIPIENTS]
        try:
            self.connection.sendmail(
                from_email, recipients, email_message.message().as_string())
        except:
            if not self.fail_silently:
                raise
            return False
        return True


class WhitelistEmailBackend(SmtpEmailBackend):
    """
    Email backend that sends only these emails, that match the whitelist
    setting.

    In order to use it, set this in your local_settings.py::

        EMAIL_BACKEND = 'django_libs.test_email_backend.EmailBackend'
        EMAIL_BACKEND_WHITELIST = [
            r'.*@example\.com',
        ]

    This setting would allow all emails to @example.com to be sent and all
    others are discarded. The setting expects regex, so better test it before
    adding it here to prevent errors.

    If the setting does not exist, no emails are sent at all.

    """
    def _send(self, email_message):
        """A helper method that does the actual sending."""
        from_email = sanitize_address(
            email_message.from_email, email_message.encoding)
        recipients = self.clean_recipients(email_message)

        if not recipients:
            return False

        try:
            self.connection.sendmail(
                from_email, recipients, email_message.message().as_string())
        except:
            if not self.fail_silently:
                raise
            return False
        return True

    def clean_recipients(self, email_message):
        """Removes all the unallowed recipients."""
        new_recipients = []

        recipients = [sanitize_address(addr, email_message.encoding)
                      for addr in email_message.recipients()]
        for recipient in recipients:
            if self.matches_whitelist(recipient):
                new_recipients.append(recipient)
            elif settings.EMAIL_BACKEND_REROUTE_BLACKLIST:
                for name, addr in settings.TEST_EMAIL_BACKEND_RECIPIENTS:
                    new_recipients.append(addr)
        # remove duplicates
        new_recipients = list(set(new_recipients))
        return new_recipients

    def matches_whitelist(self, recipient):
        """Checks if the email address matches one of the whitelist entries."""
        matches = False
        for entry in settings.EMAIL_BACKEND_WHITELIST:
            if re.match(entry, recipient):
                matches = True
        return matches
